package mvd.jester.simulator;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.TreeSet;
import mvd.jester.model.DagTask;
import mvd.jester.model.Subtask;
import mvd.jester.priority.PriorityManager;
import mvd.jester.simulator.events.*;
import mvd.jester.simulator.exceptions.DeadlineMissedException;
import mvd.jester.simulator.exceptions.SchedulingException;
import mvd.jester.simulator.model.ExtendedSubtaskContext;
import mvd.jester.simulator.model.JobContext;
import mvd.jester.simulator.model.ProcessorContext;
import mvd.jester.simulator.model.ThreadContext;
import mvd.jester.simulator.model.ProcessorContext.ProcessorComparator;

public class GlobalScheduler {

    private final int nbProcessors;
    private final PriorityManager priorityManager;

    public GlobalScheduler(final PriorityManager priorityManager, final int nbProcessors) {
        this.priorityManager = priorityManager;
        this.nbProcessors = nbProcessors;
    }

    public void schedule(final Set<DagTask> taskSet, final long duration) {
        final SchedulingContext sc = new SchedulingContext(priorityManager, nbProcessors, duration);
        taskSet.forEach(t -> sc.events.offer(new TaskReleasedEvent(0, t)));
        while (!sc.events.isEmpty()) {
            final IEvent event = sc.events.poll();
            updateExecution(sc, event.getTime());
            handle(event, sc);
        }
    }

    private void handle(final IEvent event, final SchedulingContext sc) {
        event.log();
        switch (event.getType()) {
            case SUBTASK_ACTIVATED:
                handleSubtaskActivated((SubtaskActivatedEvent) event, sc);
                break;
            case SUBTASK_COMPLETED:
                handleSubtaskCompleted((SubtaskCompletedEvent) event, sc);
                break;
            case SUBTASK_CHECK_COMPLETED:
                handleSubtaskCheckCompleted((SubtaskCheckCompletedEvent) event, sc);
                break;
            case SUBTASK_PREEMPTED:
                handleSubtaskPreempted((SubtaskPreemptedEvent) event, sc);
                break;
            case SUBTASK_STARTED:
                handleSubtaskStarted((SubtaskStartedEvent) event, sc);
                break;
            case SUBTASK_STOPPED:
                handleSubtaskStopped((SubtaskStoppedEvent) event, sc);
                break;
            case JOB_ACTIVATED:
                handleJobActivated((JobActivatedEvent) event, sc);
                break;
            case JOB_COMPLETED:
                handleJobCompleted((JobCompletedEvent) event, sc);
                break;
            case TASK_RELEASED:
                handleTaskReleased((TaskReleasedEvent) event, sc);
                break;
            default:
                throw new IllegalArgumentException(
                        event.getType() + " is not handled by " + this.getClass());
        }
    }

    private void handleJobCompleted(final JobCompletedEvent event, final SchedulingContext sc) {
        JobContext job = event.getJobContext();
        if (event.getTime() > job.getDeadline()) {
            throw new DeadlineMissedException(event.getJobContext() + " misses its deadline");
        }
        // TODO: Remove threads from sc.threadqueue?
        if (sc.threadQueue.stream().anyMatch(t -> t.getJobContext() == job)) {
            throw new RuntimeException(
                    "Why is there still a thread of this completed job in the queue?");
        }
    }

    private void handleSubtaskActivated(final SubtaskActivatedEvent event,
            final SchedulingContext sc) {
        final Subtask subtask = event.getSubtask();
        final JobContext job = event.getJobContext();
        ExtendedSubtaskContext context = new ExtendedSubtaskContext(job, subtask);
        if (job.claimThread(context)) {
            sc.threadQueue.offer(context.getThread().get());
        }
        updateAdmission(sc, event.getTime());
    }

    private void handleSubtaskCheckCompleted(final SubtaskCheckCompletedEvent event,
            final SchedulingContext sc) {
        final ProcessorContext processor = event.getProcessor();
        if (processor.isIdle()) {
            throw new SchedulingException(
                    processor + " should execute " + event.getSubtaskContext());
        }
        final Optional<ExtendedSubtaskContext> context = processor.getSubtask();
        if (context.isPresent() && context.get() == event.getSubtaskContext()
                && context.get().getRemainingExecutionTime() == 0) {
            sc.events.offer(new SubtaskStoppedEvent(event.getTime(), context.get(), processor));
        }
    }

    private void handleSubtaskCompleted(final SubtaskCompletedEvent event,
            final SchedulingContext sc) {
        final ExtendedSubtaskContext subtask = event.getSubtaskContext();
        final JobContext job = subtask.getJobContext();

        boolean isSink = job.registerFinishedSubtask(subtask);
        if (isSink) {
            sc.events.offer(new JobCompletedEvent(event.getTime(), job));
            updateAdmission(sc, event.getTime());
        } else {
            if (job.releaseThread(subtask)) {
                sc.threadQueue.offer(subtask.getThread().get());
                updateAdmission(sc, event.getTime());
            }
            List<Subtask> successors = job.getReadySuccessors(subtask);
            if (successors.isEmpty()) {
                updateAdmission(sc, event.getTime());
            } else {
                for (final Subtask s : successors) {
                    sc.events.offer(new SubtaskActivatedEvent(event.getTime(), job, s));
                }
            }
        }
    }

    private void handleSubtaskPreempted(final SubtaskPreemptedEvent event,
            final SchedulingContext sc) {
        final ProcessorContext processor = event.getProcessor();
        final long time = event.getTime();
        sc.events.offer(new SubtaskStoppedEvent(time, event.getSubtaskContext(), processor));
        sc.events.offer(new SubtaskStartedEvent(time, event.getNextSubtask(), processor));
    }

    private void handleSubtaskStarted(final SubtaskStartedEvent event, final SchedulingContext sc) {
        final long time = event.getTime();
        final ProcessorContext processor = event.getProcessor();
        final ExtendedSubtaskContext context = event.getSubtaskContext();
        if (processor.accept(context, time)) {
            sc.events.offer(new SubtaskCheckCompletedEvent(
                    time + context.getRemainingExecutionTime(), context, processor));
        } else {
            int brk = 0;
        }
    }

    private void handleSubtaskStopped(final SubtaskStoppedEvent event, final SchedulingContext sc) {
        final ExtendedSubtaskContext context = event.getSubtaskContext();
        final ProcessorContext processor = event.getProcessor();
        if (context.getRemainingExecutionTime() > 0) {
            sc.threadQueue.offer(context.getThread().get());
        } else {
            sc.events.offer(new SubtaskCompletedEvent(event.getTime(), context));
        }

        processor.free();
    }

    private void handleJobActivated(final JobActivatedEvent event, final SchedulingContext sc) {
        final JobContext job = event.getJobContext();

        final DagTask task = job.getTask();
        final long nextRelease = event.getTime() + task.getPeriod();
        if (nextRelease < sc.duration) {
            final JobContext nextJob = new JobContext(task, nextRelease);
            sc.events.offer(new JobActivatedEvent(nextRelease, nextJob));
        }

        sc.events.offer(new SubtaskActivatedEvent(event.getTime(), job, job.getSource()));
    }

    private void handleTaskReleased(final TaskReleasedEvent event, final SchedulingContext sc) {
        final DagTask task = event.getTask();
        final long releaseTime = event.getTime();
        final JobContext job = new JobContext(task, releaseTime);
        sc.events.offer(new JobActivatedEvent(releaseTime, job));
    }

    private void updateAdmission(final SchedulingContext sc, final long time) {
        Set<ProcessorContext> sortedProcessors =
                new TreeSet<>(new ProcessorComparator(priorityManager));
        sortedProcessors.addAll(sc.processors);
        if (sortedProcessors.size() != sc.processors.size()) {
            throw new RuntimeException("Processors wrong!"); // TODO: delete this later on
        }
        for (final ProcessorContext processor : sortedProcessors) {
            if (processor.canAccept(sc.threadQueue.peek(), priorityManager)) {
                ThreadContext threadContext = sc.threadQueue.poll();
                final ExtendedSubtaskContext context = threadContext.getSubtask().get();
                if (processor.isIdle()) {
                    sc.events.offer(new SubtaskStartedEvent(time, context, processor));
                } else {
                    sc.events.offer(new SubtaskPreemptedEvent(time, processor.getSubtask().get(),
                            processor, context));
                }
            }
        }
    }

    private void updateExecution(final SchedulingContext sc, final long time) {
        for (final ProcessorContext processor : sc.processors) {
            processor.update(time);
        }
    }

    private class SchedulingContext {

        private final EventQueue events;
        private final Set<ProcessorContext> processors;
        private final ThreadQueue threadQueue;
        private final long duration;

        private SchedulingContext(final PriorityManager priorityManager, final int nbProcessors,
                final long duration) {
            this.events = new EventQueue(priorityManager);
            this.threadQueue = new ThreadQueue(priorityManager);
            this.processors = new HashSet<>();
            for (int m = 0; m < nbProcessors; ++m) {
                processors.add(new ProcessorContext(m));
            }
            this.duration = duration;
        }

    }
}
