package mvd.jester.simulator.model;

import java.util.Comparator;
import java.util.Optional;
import mvd.jester.priority.PriorityManager;
import mvd.jester.simulator.exceptions.DeadlineMissedException;
import mvd.jester.simulator.exceptions.ExecutionOverrunException;

/**
 * Processor
 */
public class ProcessorContext {

    private final long processorId;
    private Optional<ExtendedSubtaskContext> currentSubtask;
    private long lastEventTime;

    public ProcessorContext(long processorId) {
        currentSubtask = Optional.empty();
        this.processorId = processorId;
        this.lastEventTime = 0;
    }

    public void setSubtask(ExtendedSubtaskContext subtask) {
        this.currentSubtask = Optional.ofNullable(subtask);
    }

    public boolean accept(ExtendedSubtaskContext context, long time) {
        if (isIdle()) {
            this.currentSubtask = Optional.of(context);
            lastEventTime = time;
            return true;
        }
        return false;
    }

    /**
     * @return the currentJob
     */
    public Optional<ExtendedSubtaskContext> getSubtask() {
        return currentSubtask;
    }

    public boolean canAccept(ThreadContext thread, PriorityManager priorityManager) {
        if (thread == null) {
            return false;
        }

        if (currentSubtask.isEmpty()) {
            return true;
        }

        if (priorityManager.compare(thread.getJobContext(),
                currentSubtask.get().getJobContext()) < 0) {
            return true;
        }

        return false;
    }

    public void free() {
        currentSubtask = Optional.ofNullable(null);
    }

    public boolean isIdle() {
        if (currentSubtask.isEmpty()) {
            return true;
        }

        return false;
    }


    public void update(long time) {
        if (currentSubtask.isPresent()) {
            currentSubtask.get().updateRemainingExecutionTime(time - lastEventTime);
            lastEventTime = time;
            long remainingTime = currentSubtask.get().getRemainingExecutionTime();
            if (remainingTime < 0) {
                throw new ExecutionOverrunException(currentSubtask.get().getJobContext()
                        + " overruns its execution at time " + time + ": " + remainingTime);
            }
            if (time > currentSubtask.get().getJobContext().getDeadline()) {
                throw new DeadlineMissedException(
                        currentSubtask.get().getJobContext() + " misses its deadline");
            }
        }
    }


    @Override
    public String toString() {
        return "Processor " + processorId;
    }


    public enum JobState {
        NOT_ACCEPTED, ACCEPTED, REMOVED
    }


    public static class ProcessorComparator implements Comparator<ProcessorContext> {

        private final PriorityManager priorityManager;

        public ProcessorComparator(PriorityManager priorityManager) {
            this.priorityManager = priorityManager;

        }

        @Override
        public int compare(ProcessorContext p1, ProcessorContext p2) {
            if (!p1.getSubtask().isPresent()) {
                return -1;
            } else if (!p2.getSubtask().isPresent()) {
                return 1;
            } else {
                JobContext p1Job = p1.currentSubtask.get().getJobContext();
                JobContext p2Job = p2.currentSubtask.get().getJobContext();
                if (p1Job == p2Job) {
                    return 1;
                }
                // Sort in reverse
                return priorityManager.compare(p2.currentSubtask.get().getJobContext(),
                        p1.currentSubtask.get().getJobContext());
            }
        }
    }

}
